from __future__ import annotations

import logging

from src.csvlineparser.char_tokenizer_interface import ICharTokenizer
from src.csvlineparser.csv_consumer_interface import ICsvConsumer
from src.csvlineparser.exceptions import CsvParserTooMuchDataException, CsvParserNoTermQuoteException
from src.csvlineparser.multi_consumer import MultiConsumer

logger = logging.getLogger(__name__)


class CsvParser:

    def __init__(self, debugOffset: int = 20) -> None:
        super().__init__()
        self.__firstRow: bool = True
        self.__currentRow: str = ''
        self.__debugOffset: int = debugOffset
        self.__lineNumber: int = 0

    def parse(self, reader: ICharTokenizer,
              consumers: ICsvConsumer | list[ICsvConsumer],
              failFast: bool = True) -> None:
        multiConsumer = MultiConsumer(failFast=failFast)
        multiConsumer.addConsumers(consumers)
        self.__parseCsvFile(reader, multiConsumer)

    def __parseCsvFile(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        while reader.peek() != 'EOF':
            self.__parseCsvRecord(reader, consumer)
        consumer.signalEndOfFile()

    def __parseCsvRecord(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        self.__lineNumber += 1
        self.__parseCsvStringList(reader, consumer)
        ch: str = reader.read()
        if ch == 'EOF':
            reader.unread(ch)
            ch = '\n'
        if ch != '\n':
            raise CsvParserTooMuchDataException(f'End of record was expected but more data exists: >{ch}<',
                                                reader.getPos(), reader.getSegment(reader.getPos()
                                                                                   - self.__debugOffset,
                                                                                   self.__debugOffset * 2),
                                                self.__debugOffset, self.__lineNumber)
        consumer.signalEndOfRecord()

    def __parseCsvStringList(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        ch: str
        while True:
            self.__parseRawString(reader, consumer)
            ch = reader.read()
            if ch != ',':
                break
        reader.unread(ch)

    def __isFieldTerminator(self, c: str) -> bool:
        return c == ',' or c == '\n' or c == 'EOF'

    def __isSpace(self, c: str) -> bool:
        return c == ' ' or c == '\t'

    def __parseOptionalSpaces(self, reader: ICharTokenizer) -> None:
        ch: str
        while True:
            ch = reader.read()
            if not self.__isSpace(ch):
                break
        reader.unread(ch)

    def __parseRawString(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        self.__parseOptionalSpaces(reader)
        self.__parseRawField(reader, consumer)
        if not self.__isFieldTerminator(reader.peek()):
            self.__parseOptionalSpaces(reader)

    def __parseRawField(self, reader: ICharTokenizer, consumer: ICsvConsumer) -> None:
        fieldValue: str = ''

        ch: str = reader.peek()
        if not self.__isFieldTerminator(ch):
            if ch == '"':
                fieldValue = self.__parseQuotedField(reader)
            else:
                fieldValue = self.__parseSimpleField(reader)
        consumer.consumeField(fieldValue.strip())

    def __parseQuotedField(self, reader: ICharTokenizer) -> str:
        reader.read()  # read and discard initial quote

        field: str = self.__parseEscapedField(reader)

        ch: str = reader.read()
        if ch != '"':
            reader.unread(ch)
            raise CsvParserNoTermQuoteException('Quoted field has no terminating double quote')
        return field

    def __parseEscapedField(self, reader: ICharTokenizer) -> str:
        sb: list[str] = list()

        self.__parseSubField(reader, sb)
        ch: str = reader.read()
        while self.__processDoubleQuote(reader, ch):
            sb.append('"')
            self.__parseSubField(reader, sb)
            ch = reader.read()
        reader.unread(ch)

        return ''.join(sb)

    def __parseSubField(self, reader: ICharTokenizer, sb: list[str]) -> None:
        ch: str = reader.read()
        while ch != '"' and ch != 'EOF':
            sb.append(ch)
            ch = reader.read()
        reader.unread(ch)

    def __isBadSimpleFieldChar(self, c: str) -> bool:
        return self.__isFieldTerminator(c) or c == '"'

    def __parseSimpleField(self, reader: ICharTokenizer) -> str:
        ch: str = reader.read()
        if self.__isBadSimpleFieldChar(ch) or self.__isSpace(ch):
            reader.unread(ch)
            return ''

        sb: list[str] = list()
        sb.append(ch)
        ch = reader.read()
        while not self.__isBadSimpleFieldChar(ch):
            sb.append(ch)
            ch = reader.read()
        reader.unread(ch)
        return ''.join(sb)

    def __processDoubleQuote(self, reader: ICharTokenizer, ch: str) -> bool:
        if ch == '"' and reader.peek() == '"':
            reader.read()  # discard second quote of double
            return True
        return False



from __future__ import annotations

import logging

from src.csvlineparser.csv_consumer_interface import ICsvConsumer

logger = logging.getLogger(__name__)


class MultiConsumer(ICsvConsumer):

    def __init__(self, failFast: bool = False) -> None:
        """ ICsvConsumer can be appended, and it will delegate the call to many.
        If failFast, will raise an exception data columns differ from header columns
        (or generally the first column) """
        super().__init__()
        self.__headerRead: bool = False
        self.__headerElemCount = 0
        self.__lineElemCount = 0
        self.__consumers: list[ICsvConsumer] = []
        self.__failFast: bool = failFast
        self.__lineNum = 1

    def addConsumers(self, consumer: ICsvConsumer | list[ICsvConsumer]) -> None:
        if isinstance(consumer, list):
            self.__consumers.extend(consumer)
        else:
            self.__consumers.append(consumer)

    def consumeField(self, s: str) -> None:
        if not self.__headerRead:
            self.__headerElemCount += 1
        else:
            self.__lineElemCount += 1
        for consumer in self.__consumers:
            consumer.consumeField(s)

    def signalEndOfRecord(self) -> None:
        if self.__headerRead:
            if self.__headerElemCount != self.__lineElemCount:
                if self.__failFast:
                    raise Exception(f'Header element count [{self.__headerElemCount}] does not '
                                    f'match line element count [{self.__lineElemCount}]. '
                                    f'Line: {self.__lineNum}')
        self.__headerRead = True
        self.__lineElemCount = 0
        self.__lineNum += 1
        for consumer in self.__consumers:
            consumer.signalEndOfRecord()

    def signalEndOfFile(self) -> None:
        for consumer in self.__consumers:
            consumer.signalEndOfFile()



import logging

from src.csvlineparser.csv_consumer_interface import ICsvConsumer

logger = logging.getLogger(__name__)


class CsvConsumerTester(ICsvConsumer):

    def __init__(self, expectedFields: list[str]) -> None:
        super().__init__()
        self.expectedFields: list[str] = expectedFields
        self.fieldIndex: int = 0

    def consumeField(self, s: str) -> None:
        if self.expectedFields[self.fieldIndex] != s:
            raise Exception(f'field [{self.expectedFields[self.fieldIndex]}] expected, but [{s}] returned')
        self.fieldIndex += 1

    def signalEndOfRecord(self) -> None:
        if self.expectedFields[self.fieldIndex] != 'EOR':
            self.fieldIndex += 1
            raise Exception('End of record signalled but not expected')

    def signalEndOfFile(self) -> None:
        if self.expectedFields[self.fieldIndex] != '':
            self.fieldIndex += 1
            raise Exception('End of file signalled but not expected')



import logging

from src.csvlineparser.csv_consumer_interface import ICsvConsumer

logger = logging.getLogger(__name__)


class CsvConsoleConsumer(ICsvConsumer):

    def __init__(self) -> None:
        super().__init__()
        self.__colIndex = 0
        self.__rowIndex = 0
        self.__numCols = 0
        self.__numRows = 0
        self.__matrix = {}

    def consumeField(self, s: str) -> None:
        self.__matrix[self.__colIndex, self.__rowIndex] = s
        self.__colIndex += 1
        if self.__numCols < self.__colIndex:
            self.__numCols += 1

    def signalEndOfRecord(self) -> None:
        self.__rowIndex += 1
        self.__numRows += 1
        self.__colIndex = 0

    def signalEndOfFile(self) -> None:
        print(f'End of file reached. Got {self.__numCols} columns and {self.__numRows} rows')

    def getMatrix(self) -> dict:
        return self.__matrix

    def getColumn(self, colIndex: int) -> list:
        col = []
        for i in range(0, self.__numRows):
            col.append(self.__matrix[colIndex, i])
        return col

    def getRow(self, rowIndex: int) -> list:
        row = []
        for i in range(0, self.__numCols):
            row.append(self.__matrix[i, rowIndex])
        return row

    def getNumRows(self) -> int:
        return self.__numRows

    def geNumCols(self) -> int:
        return self.__numCols



from __future__ import annotations

import logging

from src.csvlineparser.csv_console_consumer import CsvConsoleConsumer
from src.csvlineparser.csv_context_provider import CsvContextProvider

logger = logging.getLogger(__name__)


class CsvCleanerConsumer(CsvConsoleConsumer):

    def __init__(self, hasHeader: bool = True, csvContextProvider: CsvContextProvider | None = None) -> None:
        super().__init__()
        self.csvContextProvider = csvContextProvider
        if csvContextProvider:
            self.delimiter = csvContextProvider.getDelimiter()
        else:
            self.delimiter = ','
        self.hasHeader = hasHeader
        self.__line = ''
        self.__lineNum = 0
        self.__colCount = 0
        self.__allLines = []

    def consumeField(self, inputValue: str) -> None:
        super().consumeField(inputValue)

        if self.csvContextProvider:
            if self.__lineNum == 0:
                if self.hasHeader:
                    inputValue = self.csvContextProvider.handleHeaderElement(inputValue, self.__lineNum,
                                                                             self.__colCount)
                else:
                    inputValue = self.csvContextProvider.handleDataElement(inputValue, self.__lineNum, self.__colCount)
            else:
                inputValue = self.csvContextProvider.handleDataElement(inputValue, self.__lineNum, self.__colCount)

        self.__line += inputValue + self.delimiter
        self.__colCount += 1

    def signalEndOfRecord(self) -> None:
        super().signalEndOfRecord()
        self.__line = self.__line[:-1]
        self.__allLines.append(self.__line)
        self.__line = ''
        self.__lineNum += 1
        self.__colCount = 0

    def signalEndOfFile(self) -> None:
        super().signalEndOfFile()

    def getHeader(self) -> str:
        return self.__allLines[0]

    def getData(self) -> list:
        return self.__allLines[1:]

    def getAll(self) -> list:
        return self.__allLines

    def getLine(self, lineNum: int) -> str:
        return self.__allLines[lineNum]
